/* 
 * FILE --- mlp_lm.hh
 *
 * PURPOSE --- Header file for a class that implements a Multilayer 
 * Perceptron (MLP) neural network (NN) with Levenberg-Marquardt (LM) training capability. 
 * The user specifies the number of inputs, the number of outputs, and the number of the 
 * neurons in the hidden layer of the perceptron as well as the number data points that 
 * will be used for training. The traiing algorithm operates in a batch mode, therefore,
 * the user has to provide all the data at the beginning of the training process.
 * It is possible to choose two different activation functions for the neurons that are
 * Log-Sigmoid and Hyperbolic Tangent Sigmoid. The user can either specify the initial 
 * input and output weigth matrices and biaces, or let them be choosen randomly within 
 * user defined range.
 *
 * This utility uses the matrix tool for matrix maltiplications.
 *
 */


// Prevent Multiple Inclusion
#ifndef MLP_LM_HH
#define MLP_LM_HH


// Include Files
#include <stdlib.h>
#include "matrix.hh"

// Local Definitions and Macros
#define MLP_LM_NO_ERROR 0
#define MLP_LM_MEM_ERROR 1

#define LOGSIG 0  // Log-Sigmoid activation function
#define TANSIG 1  // Hyperbolic Tangent Sigmoid activation function

#define NORM_LOWER_BOUND 0.001   // Lower bound on the norm of the gradient that will
                                 // cause the algorithm to stop
#define LAMBDA_UPPER_BOUND 100000.0  // Dont let lambda to be bigger than this value
#define LAMBDA_LOWER_BOUND 0.00001   // Dont let lambda to be smaller than this value

class MLP_LM
{
public:
  
  // Constructor 
  MLP_LM(int _num_of_inputs, int _num_of_outputs, int _num_of_neurons, int _num_of_data);

  // Destructor
  ~MLP_LM();

  // Functions to set the input weights and biases
  void set_input_gains(double _low, double _high);
  void set_input_gains(const double **_in_gains);
  void set_input_gains(const double *_in_gains);
  void set_input_biases(double _low, double _high);
  void set_input_biases(const double *_in_biases);

  // Functions to set the output weights and biases
  void set_output_gains(double _low, double _high);
  void set_output_gains(const double **_out_gains);
  void set_output_gains(const double *_out_gains);
  void set_output_biases(double _low, double _high);
  void set_output_biases(const double *_out_biases);

  // Function to set the parameter lambda used to prevent inversion
  // of singular matrices (and to define the step size)
  void set_lambda(double _lambda);

  // Set or return the type of the activation functions
  void set_activation_function_type(int _actv_func_type);
  int get_activation_function_type(void) { return actv_func_type; }

  // The function that given the inputs, desired outputs and the tolerance in the sum of 
  // the error trains the network (i.e., updates the parameters to obtain the desired
  // tolerance). The function operates in a batch mode, therefore, the user has to provide
  // all the training data to the function.
  void train(const double **input, const double **ref_output, double _tol);

  // Functions that performs only one step of the above algorithm
  // Returns the current error
  double train_one_step(const double **input, const double **ref_output);

  // The function that can be used to test the network.
  // Given the input vector it computes and returns the network output  
  void test(const double *input, double *output);
  
  // Function that returns the current norm of the gradient
  inline double get_norm_grad(void);

  int mlp_lm_err; // Variable to hold the error condition

private:
  // Allocate/free memory for the arrays
  void allocate_memory(void);
  void free_memory(void);

  // Function to transfer the parameters
  void Wi2theta(void);
  void bi2theta(void);
  void Wo2theta(void);
  void bo2theta(void);

  void theta2Wi(void);
  void theta2bi(void);
  void theta2Wo(void);
  void theta2bo(void);

  // Function that calculates the activation of each neuron
  void get_activation(void);

 // Function that transfers the outputs from a matrix to the output variable
  void get_output(double *_y);

  // Function that calculates the activation of each neuron for all data
  void get_F(void);

  // Function that is used to add the output bias term to all data
  void add_out_bias(void);

  // Function that calculates the derivative of each activation function for all data
  void get_del_F(void);

   // Function that calculates the Jacobian
  void get_J(void);

 // Function that calculates the vector of output errors
  void get_out_error(void);

  // Pointer to the currently chosen activation function for the neurons
  double (*actv_func) (const double input);

  // Pointer to the currently chosen activation function derivative
  double (*del_actv_func) (const double input);

  int num_of_inputs;  // # of inputs (to be specified by the user)
  int num_of_outputs;  // # of outputs (to be specified by the user)
  int num_of_neurons;  // # of neurons (to be specified by the user)
  int num_of_data;  // # of data points (to be specified by the user)

  int num_of_params;  // total # of parameters
  int index1;  // boundary of the 1. partition in the J and theta
  int index2;  // boundary of the 2. partition in the J and theta
  int index3;  // boundary of the 3. partition in the J and theta

  int actv_func_type;  // Type of activation function

  double lambda;  // Parameter that is used to prevent inversion of singular matrices 
                  // (and to define the step size)

  // The weigth and bias matrices
  matrix *Wi;  // Input weigth matrix
  matrix *bi;  // Input bias matrix
  matrix *Wo;  // Output weigth matrix
  matrix *bo;  // Output bias matrix

  // Convaniance variables
  matrix *u; // Input vector
  matrix *x; // Intermediate vector
  matrix *f; // Intermediate vector
  matrix *y; // Output vector

  // Varaibles needed for adaptation
  matrix *theta;  // Vector of all parameters
  matrix *theta_old;  // Vector of all parameters
  matrix *J;  // Jacobian matrix
  matrix *U; // Input matrix
  matrix *X; // Intermediate matrix
  matrix *Y; // Output matrix
  matrix *R; // Desired output vector
  matrix *E; // Vector of output Errors
  matrix *E_old; // Vector of output Errors
  matrix *F; // Intermediate matrix
  matrix *del_F; // Derivative of the activation functions
  matrix *Lambda;  // Diagonal matrix (lambda*I) to prevent inversion of singular matrices

};

// Activation Functions for the neurons and their derivatives wrt input parameters
inline double logsig(const double _input);
inline double tansig(const double _input);
inline double del_logsig(const double _input);
inline double del_tansig(const double _input);

#endif // MLP_LM_HH
