/* 
 * FILE --- mlp_ex.cc
 *
 * PURPOSE --- To provide an example use of the class for 4'th
 * order Runge-Kutta simulation, RK4_SIM, and the class for implementing
 * n-input m-output Multilayer Perceptron (MLP) neural network (NN).
 *
 * Implements the main function of a class for performing a simulation of
 * a simple cargo ship whose steering is controlled by a MLP controller.
 *
 */

// Include Files
#include <stdlib.h>
#include <stdio.h>

#include "mlp_lm.hh"
#include "random.hh"

// Local Definitions and Macros
#define PI 3.14159265359
#define NUM_DATA 1000
#define NUM_TEST 30

void main(void) 
{
  int ii, jj;  // Index variales for the loops
  double max_val = 2.5;             // Sampling time

  // Create the MLP object
  MLP_LM my_mlp_ctrl(2, 2, 30, NUM_DATA);

  double **input_data;      // Control variable
  double **desired_output;  // Reference input variable
  double test_input[2], test_output[2], desired_test_output[2];

  FILE *fp;  // To save data

  // Set the initial value of lambda
  my_mlp_ctrl.set_lambda(0.1);

  // Set the activation functions to be Log-Sigmoids
  // my_mlp_ctrl.set_activation_function_type(LOGSIG);
  my_mlp_ctrl.set_activation_function_type(TANSIG);

  // Create the input data matrix
  input_data = new double*[2];
  input_data[0] = new double[NUM_DATA];
  input_data[1] = new double[NUM_DATA];

  // Create the output data matrix
  desired_output = new double*[2];
  desired_output[0] = new double[NUM_DATA];
  desired_output[1] = new double[NUM_DATA];

  // Generate the training data
  for(ii = 0; ii < NUM_DATA; ii++)
    {
      // Generate the training value randomly
      input_data[0][ii] = Rand_Double(-max_val, max_val);
      input_data[1][ii] = Rand_Double(-max_val, max_val);

      // Generate the desired output
      desired_output[0][ii] = sin(input_data[0][ii])*cos(input_data[1][ii]);
      desired_output[1][ii] = sin(input_data[0][ii]) + cos(input_data[1][ii]);
    }

  // Train the network
  my_mlp_ctrl.train((const double **)input_data, (const double **)desired_output, 0.001);

  if (!(fp = fopen("trained_mlp_out.dat","w"))) {
    printf("Can't open file.\n");
    exit(1);
  }

  // Test the trained network
  test_input[0] = -max_val;
  for(ii = 0; ii < NUM_TEST; ii++)
    {
      test_input[1] = -max_val;
      for(jj = 0; jj < NUM_TEST; jj++)
	{
	  // Generate the desired output
	  desired_test_output[0] = sin(test_input[0])*cos(test_input[1]);
	  desired_test_output[1] = sin(test_input[0]) + cos(test_input[1]);
	  
	  // Test the network for the generated data
	  my_mlp_ctrl.test(test_input, test_output);
      
	  // Save the data
	  fprintf(fp,"%f %f %f %f %f %f \n", 
		  test_input[0], test_input[1],
		  desired_test_output[0], desired_test_output[1], 
		  test_output[0], test_output[1]);

	  // Move toi new test data
	  test_input[1] += 2.0*max_val/(1.0*(NUM_TEST - 1));
	}
      // Move toi new test data
      test_input[0] += 2.0*max_val/(1.0*(NUM_TEST - 1));
    }
  fclose(fp);

  delete [] input_data[0];
  delete [] input_data[1];
  delete [] input_data;
  delete [] desired_output[0];
  delete [] desired_output[1];
  delete [] desired_output;

  printf("Simulation completed ...\n");
}
